(* file: overlap.ml
 * author: Bob Muller
 * date: November 17, 2015
 *
 * This is an OCaml program solving the Overlap Graphs problem.
 * See: http://rosalind.info/problems/grph/
 *
 * Compile:
 * 
 * > ocamlc -o overlap overlap.ml
 *
 * Run:
 * 
 * > ./overlap file.fas N
 *
 * where file.fas is a FASTA file and N is an integer specifying the
 * overlap length. The program will generate a file file.graph showing
 * the overlap graph.
 *)
type id  = string
type seq = string
type fasta = (id * seq) list

(* seqsMatch : seq -> seq -> int -> bool
 *
 * The call (seqsMatch seq1 seq2 n) returns true if n-length suffix if
 * seq1 is identical to the n-length prefix of seq2.
 *)
let seqsMatch (seq1 : seq) (seq2 : seq) (n : int) : bool =
  let suffix = String.sub seq1 (String.length seq1 - n) n in
  let prefix = String.sub seq2 0 n
  in
  suffix = prefix

(* matchAll : seq -> fasta -> int -> id list
 *
 * The call (matchAll seq fasta n) returns the list of ids of
 * sequences in fasta that match seq.
 *)
let rec matchAll (seq : seq) (fasta : fasta) (n : int) =
  match fasta with
  | [] -> []
  | (id, seq')::fasta -> let almost = matchAll seq fasta n
			 in
			 match (seq <> seq') && (seqsMatch seq seq' n) with
			 | true  -> id::almost
			 | false -> almost

(* readFasta : string -> fasta
 *
 * The call (readFasta filename) opens filename and returns 
 * a list if id * sequence pairs.
 *)
let readFasta (file : string) : fasta =
  let inch = open_in file in
  let rec repeat fasta =
    try
      let id = input_line inch in
      let seq = input_line inch
      in
      repeat ((id, seq)::fasta)
    with
      End_of_file -> close_in inch;
		     fasta
  in
  repeat []
		     
(* overlapGraph : string -> int -> (id * id list) list 
 *
 * The call (overlapGraph fileName n) opens a fasta file and
 * computes the overlap graph if the sequences in the file.
 *)
let overlapGraph (fileName : string) (n : int) : (id * id list) list =
  let (fasta : fasta) = readFasta fileName 
  in
  List.map (fun (id, seq) -> (id, matchAll seq fasta n)) fasta
      
let makeFileName fname ext =
  String.sub fname 0 ((String.index fname '.') + 1) ^ ext
		
let writeGraph graph file =
  let ouch = open_out file in
  let rec repeat graph =
    match graph with
    | [] -> close_out ouch
    | (id, ids)::graph ->
       List.iter (fun id' -> output_string ouch (id ^ " -> " ^ id' ^ "\n")) ids;
       repeat graph
  in
  repeat graph

let () =
  let inFileName = Sys.argv.(1) in
  let n = int_of_string (Sys.argv.(2)) in
  let graph : (id * id list) list = overlapGraph inFileName n in
  let outFileName = makeFileName inFileName "graph"
  in
  writeGraph graph outFileName
 
